use std::{io::Write, path::PathBuf};

use actix_web::{
    HttpResponse, Responder, get, put,
    web::{Data, Path, Payload},
};
use actix_web_httpauth::extractors::bearer::BearerAuth;
use futures_util::StreamExt;
use tracing::{debug, error, trace};

use crate::{
    access::{AccessLevel, has_access_level},
    config::AppConfig,
};

fn get_cache_entry_path(config: &AppConfig, hash: &str) -> PathBuf {
    PathBuf::from(&config.cache_dir).join(hash)
}

#[get("/v1/cache/{hash}")]
pub async fn get_cache_item(
    app_data: Data<AppConfig>,
    auth: BearerAuth,
    hash: Path<String>,
) -> impl Responder {
    trace!("Requested cache item {}", hash.as_str());
    if !has_access_level(&app_data, auth.token(), AccessLevel::Read) {
        debug!(
            "Tried to read cache item {} without valid read access token.",
            hash.as_str()
        );
        return HttpResponse::Unauthorized()
            .content_type("text/plain")
            .body("Please provide a valid access token with at least read-level access.");
    }
    let path = get_cache_entry_path(&app_data, &hash);
    if !path.exists() {
        trace!("Cache item not found: {}", hash.as_str());
        HttpResponse::NotFound().body("The record was not found.")
    } else {
        trace!("Returning cache item {}", hash.as_str());
        HttpResponse::Ok().body(std::fs::read(path).unwrap())
    }
}

#[put("/v1/cache/{hash}")]
pub async fn put_cache_item(
    app_data: Data<AppConfig>,
    auth: BearerAuth,
    hash: Path<String>,
    mut body: Payload,
) -> impl Responder {
    trace!("Received cache item {}", hash.as_str());
    if !has_access_level(&app_data, auth.token(), AccessLevel::ReadWrite) {
        debug!(
            "Tried to write cache item {} without valid read-write access token.",
            hash.as_str()
        );
        return HttpResponse::Unauthorized()
            .content_type("text/plain")
            .body("Please provide a valid access token with read-write access.");
    }
    let path = get_cache_entry_path(&app_data, &hash);
    let file = std::fs::File::create_new(&path);
    match file {
        Ok(mut file) => {
            while let Some(chunk) = body.next().await {
                match chunk {
                    Ok(chunk) => file.write_all(&chunk).expect("This should actually work"),
                    Err(e) => {
                        error!("Could not write cache item chunk: {}", e);
                        drop(file);
                        std::fs::remove_file(path).unwrap(); // Clean up to make sure the block doesn't get half-written with the wrong content
                        break;
                    }
                }
            }
            debug!("Created cache item {}", hash.as_str());
            HttpResponse::Accepted().finish()
        }
        Err(_) => {
            trace!("Tried to overwrite existing cache item {}", hash.as_str());
            HttpResponse::Conflict().body("Cannot overwrite an existing record.")
        }
    }
}
